from typing import Any
from typing import Dict
from typing import Optional
from typing import Self
from typing import TypeVar

from openinference.semconv.trace import OpenInferenceSpanKindValues
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field

from grafi.common.models.default_id import default_id
from grafi.common.models.execution_context import ExecutionContext
from grafi.common.models.message import Messages
from grafi.common.models.message import MsgsAGen


class Tool(BaseModel):
    """
    A base class representing a tool in the agent.

    This class defines the interface for all tools. Each specific tool should
    inherit from this class and implement its methods.
    """

    tool_id: str = default_id
    name: Optional[str] = Field(default=None)
    type: Optional[str] = Field(default=None)
    oi_span_type: OpenInferenceSpanKindValues

    model_config = ConfigDict(arbitrary_types_allowed=True)

    class Builder:
        """Inner builder class for Tool construction."""

        def __init__(self) -> None:
            self._tool = self._init_tool()

        def _init_tool(self) -> "Tool":
            return Tool.model_construct()

        def name(self, name: str) -> Self:
            self._tool.name = name
            return self

        def type(self, type_name: str) -> Self:
            self._tool.type = type_name
            return self

        def oi_span_type(self, oi_span_type: OpenInferenceSpanKindValues) -> Self:
            self._tool.oi_span_type = oi_span_type
            return self

        def build(self) -> "Tool":
            return self._tool

    def execute(
        self,
        execution_context: ExecutionContext,
        input_data: Messages,
    ) -> Messages:
        """
        Process the input data and return a response.

        This method should be implemented by all subclasses to define
        the specific behavior of each tool.

        Raises:
            NotImplementedError: If the method is not implemented by a subclass.
        """
        raise NotImplementedError("Subclasses must implement this method.")

    async def a_execute(
        self,
        execution_context: ExecutionContext,
        input_data: Messages,
    ) -> MsgsAGen:
        yield []  # Too keep mypy happy
        raise NotImplementedError("Subclasses must implement this method.")

    def to_messages(self, response: Any) -> Messages:
        """
        Convert the tool's response to a Message object.

        Args:
            response (Any): The response generated by the tool.

        Returns:
            Message: The response converted to a Message object.

        Raises:
            NotImplementedError: If the method is not implemented by a subclass.
        """
        raise NotImplementedError("Subclasses must implement this method.")

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert the tool instance to a dictionary representation.

        Returns:
            Dict[str, Any]: A dictionary representation of the tool.
        """
        return {
            "tool_id": self.tool_id,
        }


T = TypeVar("T", bound="Tool")  # the Tool subclass
